/**
  ******************************************************************************
  * @file           : flash.h
  * @brief          : Header for flash.c file.
  *                   
  ******************************************************************************
  * @attention
  *
  ******************************************************************************
  */

/* Define to prevent recursive inclusion -------------------------------------*/
#ifndef __FLASH_H
#define __FLASH_H

/* Includes ------------------------------------------------------------------*/

/* Defines -------------------------------------------------------------------*/

/*
 * Standard SPI-NAND flash commands
 */
#define SPINAND_CMD_RESET			            0xff
#define SPINAND_CMD_GET_FEATURE			        0x0f
#define SPINAND_CMD_SET_FEATURE			        0x1f
#define SPINAND_CMD_PAGE_READ			        0x13
#define SPINAND_CMD_READ_PAGE_CACHE_RDM		  	0x30
#define SPINAND_CMD_READ_PAGE_CACHE_LAST	  	0x3f
#define SPINAND_CMD_READ_FROM_CACHE		      	0x03
#define SPINAND_CMD_READ_FROM_CACHE_FAST	  	0x0b
#define SPINAND_CMD_READ_FROM_CACHE_X2		  	0x3b
#define SPINAND_CMD_READ_FROM_CACHE_DUAL_IO		0xbb
#define SPINAND_CMD_READ_FROM_CACHE_X4		  	0x6b
#define SPINAND_CMD_READ_FROM_CACHE_QUAD_IO		0xeb
#define SPINAND_CMD_BLK_ERASE			        0xd8
#define SPINAND_CMD_PROG_EXC			        0x10
#define SPINAND_CMD_PROG_LOAD			        0x02
#define SPINAND_CMD_PROG_LOAD_RDM_DATA		   	0x84
#define SPINAND_CMD_PROG_LOAD_X4		        0x32
#define SPINAND_CMD_PROG_LOAD_RDM_DATA_X4	  	0x34
#define SPINAND_CMD_PROG_LOAD_RDM_QUAD			0x72
#define SPINAND_CMD_READ_ID			            0x9f
#define SPINAND_CMD_WR_DISABLE			        0x04
#define SPINAND_CMD_WR_ENABLE			        0x06
#define SPINAND_CMD_END				            0x00


/* feature registers */
#define REG_BLOCK_LOCK		  0xa0
#define REG_CFG			        0xb0
#define REG_STATUS		      0xc0

/* status */
#define STATUS_OIP_MASK		  0x01
#define STATUS_READY		   (0 << 0)
#define STATUS_BUSY		     (1 << 0)

#define STATUS_E_FAIL_MASK	0x04
#define STATUS_E_FAIL		   (1 << 2)

#define STATUS_P_FAIL_MASK	0x08
#define STATUS_P_FAIL		   (1 << 3)


/*Configuration register defines*/
#define CFG_ECC_MASK		    0X10
#define CFG_ECC_ENABLE	    0x10
#define CFG_OTP_MASK		    0xc2
#define CFG_OTP_ENTER		    0x40
#define CFG_OTP_EXIT		    0x00
#define CFG_OTP_PROTECT	    0xc0
#define CFG_SNOR_ENABLE	    0x82

/* block lock */
#define BL_LOCKED_MASK			0x3c
#define BL_ALL_LOCKED		    0x38
#define BL_U_1_64_LOCKED		0x08
#define BL_U_1_32_LOCKED		0x10
#define BL_U_1_16_LOCKED		0x18
#define BL_U_1_8_LOCKED		  0x20
#define BL_U_1_4_LOCKED		  0x21
#define BL_U_1_2_LOCKED		  0x30
#define BL_L_1_64_LOCKED		0x0c
#define BL_L_1_32_LOCKED		0x14
#define BL_L_1_16_LOCKED		0x1c
#define BL_L_1_8_LOCKED		  0x24
#define BL_L_1_4_LOCKED		  0x2c
#define BL_L_1_2_LOCKED		  0x34
#define BL_ALL_UNLOCKED		  0x00

#define SPI_NAND_AS5F_ECC_MASK            0x30
#define SPI_NAND_AS5F_ECC_MAX_BIT         0x30
#define SPI_NAND_AS5F_ECC_SHIFT           0x10
#define SPI_NAND_AS5F_ECC_0_BIT           0x00
#define SPI_NAND_AS5F_ECC_UNCORR          0x20

struct alliance_onfi_specific {
	uint8_t		two_plane_page_read;		/*166*/
	uint8_t		reserved0[8];			/*167-174*/
	uint8_t		otp_mode;				/*175*/
	uint8_t		otp_page_start;		/*176*/
	uint8_t		otp_data_protect_addr;	/*177*/
	uint8_t		otp_page_number;			/*178*/
	uint8_t		otp_feature_addr;			/*179*/
	uint8_t		reserved1[68];			/*180-247*/
	uint8_t		ecc_ability;			/*248*/
	uint8_t		die_selection;			/*249*/
	uint8_t		reserved2[3];			/*250-252*/
	uint8_t		parameter_page_version;	/*253*/
}__attribute__((packed));

union spi_nand_vendor_specific {
	uint8_t		vendor_specific[88];
	struct alliance_onfi_specific alliance_sepcific;
};

struct spi_nand_onfi_params {
	/* rev info and features block */
	/* 'O' 'N' 'F' 'I'  */
	uint8_t		sig[4];				/*0-3*/
	uint16_t		revision;			/*4-5*/
	uint16_t		features;			/*6-7*/
	uint16_t		opt_cmd;			/*8-9*/
	uint8_t		reserved0[22];			/*10-31*/

	/* manufacturer information block */
	char		manufacturer[12];		/*32-43*/
	char		model[20];			/*44-63*/
	uint8_t		mfr_id;				/*64*/
	uint16_t		date_code;			/*65-66*/
	uint8_t		reserved1[13];			/*67-79*/

	/* memory organization block */
	uint32_t		byte_per_page;			/*80-83*/
	uint16_t		spare_bytes_per_page;		/*84*85*/
	uint32_t		data_bytes_per_ppage;		/*86-89*/
	uint16_t		spare_bytes_per_ppage;		/*90-91*/
	uint32_t		pages_per_block;		/*92-95*/
	uint32_t		blocks_per_lun;			/*96-99*/
	uint8_t		lun_count;			/*100*/
	uint8_t		addr_cycles;			/*101*/
	uint8_t		bits_per_cell;			/*102*/
	uint16_t		bb_per_lun;			/*103-104*/
	uint16_t		block_endurance;		/*105-106*/
	uint8_t		guaranteed_good_blocks;		/*107*/
	uint16_t		guaranteed_block_endurance;	/*108-109*/
	uint8_t		programs_per_page;		/*110*/
	uint8_t		ppage_attr;			/*111*/
	uint8_t		ecc_bits;			/*112*/
	uint8_t		interleaved_bits;		/*113*/
	uint8_t		interleaved_ops;		/*114*/
	uint8_t		reserved2[13];			/*115-127*/

	/* electrical parameter block */
	uint8_t		io_pin_capacitance_max;		/*128*/
	uint16_t		timing_mode;			/*129-130*/
	uint16_t		program_cache_timing_mode;	/*131-132*/
	uint16_t		t_prog;				/*133-134*/
	uint16_t		t_bers;				/*135-136*/
	uint16_t		t_r;				/*137-138*/
	uint16_t		t_ccs;				/*139-140*/
	uint8_t		reserved3[23];			/*141-163*/

	/* vendor */
	uint16_t		vendor_specific_revision;	/*164-165*/
	union spi_nand_vendor_specific vendor;		/*166-253*/

	uint16_t		crc;				/*254-255*/
} __attribute__((packed));

#define ONFI_CRC_BASE	0x4F4E

/**
 * struct spi_nand_chip - SPI-NAND Private Flash Chip Data
 * @name:		name of the chip
 * @spi:		[INTERN] point to spi device structure
 * @mfr_id:		[BOARDSPECIFIC] manufacture id
 * @dev_id:		[BOARDSPECIFIC] device id
 * @read_cache_op:	[REPLACEABLE] Opcode of read from cache
 * @write_cache_op:	[REPLACEABLE] Opcode of program load
 * @write_cache_rdm_op:	[REPLACEABLE] Opcode of program load random
 * @oobbuf:		[INTERN] buffer for read/write oob
 * @size:		[INTERN] the size of chip
 * @block_size:		[INTERN] the size of eraseblock
 * @page_size:		[INTERN] the size of page
 * @oob_size:	[INTERN] the size of page oob size
 * @block_shift:	[INTERN] number of address bits in a eraseblock
 * @page_shift:		[INTERN] number of address bits in a page (column
 *			address bits).
 * @pagemask:		[INTERN] page number mask = number of (pages / chip) - 1
 * @options:		[BOARDSPECIFIC] various chip options. They can partly
 *			be set to inform nand_scan about special functionality.
 * @ecc_strength:	[INTERN] ECC correctability
 * @refresh_threshold:	[INTERN] Bitflip threshold to return -EUCLEAN
 * @ecclayout:		[BOARDSPECIFIC] ECC layout control structure
 *			See the defines for further explanation.
 * @onfi_params:	[INTERN] holds the ONFI page parameter
 */
struct spi_nand_chip {
	char		*name;
	struct spi_slave	*spi;

	uint8_t		mfr_id;
	uint8_t		dev_id;
	uint8_t		read_cache_op;
	uint8_t		write_cache_op;
	uint8_t		write_cache_rdm_op;

	uint8_t		*oobbuf;
	uint64_t		size;
	uint32_t		block_size;
	uint16_t		page_size;
	uint16_t		oob_size;
	uint8_t		lun_shift;
	uint8_t		block_shift;
	uint8_t		page_shift;
	uint16_t		page_mask;
	uint32_t		options;
	uint32_t		ecc_strength;
	uint8_t		refresh_threshold;
	uint8_t		lun;
	struct nand_ecclayout *ecclayout;
	struct spi_nand_onfi_params	 onfi_params;
};

struct spi_nand_flash {
	char		*name;
	uint8_t		mfr_id;
	uint8_t		dev_id;
	uint32_t		page_size;
	uint32_t		oob_size;
	uint32_t		pages_per_blk;
	uint32_t		blks_per_lun;
	uint32_t		luns_per_chip;
	uint32_t		ecc_strength;
	uint32_t		options;
};

#define SPINAND_MAX_ADDR_LEN		4

struct spi_nand_cmd {
	uint8_t		cmd;
	uint8_t		n_addr;		/* Number of address */
	uint8_t		addr[SPINAND_MAX_ADDR_LEN];	/* Reg Offset */
	uint32_t		n_tx;		/* Number of tx bytes */
	const uint8_t	*tx_buf;	/* Tx buf */
	uint32_t		n_rx;		/* Number of rx bytes */
	uint8_t		*rx_buf;	/* Rx buf */
};

enum {
	NORMAL_MODE,
	OTP_MODE,
	OTP_PROTECT_MODE,
	SNOR_READ_ENABLE_MODE,
};

#define SPI_NAND_INFO(nm, mid, did, pagesz, oobsz, pg_per_blk,\
	blk_per_lun, lun_per_chip, ecc_stren, opts)		\
	{ .name = (nm), .mfr_id = (mid), .dev_id = (did),\
	.page_size = (pagesz), .oob_size = (oobsz),\
	.pages_per_blk = (pg_per_blk), .blks_per_lun = (blk_per_lun),\
	.luns_per_chip = (lun_per_chip),			\
	.ecc_strength = (ecc_stren), .options = (opts) }


#define SPINAND_MFR_ALLIANCE		 0x52
#define SPINAND_MAX_ID_LEN		   2

#ifdef DEBUG
#define spi_nand_debug(fmt,args...)	printk(fmt ,##args)
#else
#define spi_nand_debug(fmt,args...)
#endif
#define spi_nand_error(fmt,args...)	printk(fmt ,##args)
#define spi_nand_info(fmt,args...)	printk(fmt ,##args)

/*free oob descrption*/
struct nand_oobfree {
	uint32_t offset;
	uint32_t length;
};

/*ECC layout description*/
struct nand_ecclayout {
	uint32_t eccbytes;
	uint32_t eccpos[680];
	uint32_t oobavail;
	struct nand_oobfree oobfree[32];
};

/*oob operation mode*/
enum {
	MTD_OPS_PLACE_OOB = 0,
	MTD_OPS_AUTO_OOB = 1,
	MTD_OPS_RAW = 2,
};

/**
 * struct mtd_oob_ops - oob operation operands
 * @mode:	operation mode
 * @len:	number of data bytes to write/read
 * @retlen:	number of data bytes written/read
 * @ooblen:	number of oob bytes to write/read
 * @oobretlen:	number of oob bytes written/read
 * @ooboffs:	offset of oob data in the oob area (only relevant when
 *		mode = MTD_OPS_PLACE_OOB or MTD_OPS_RAW)
 * @datbuf:	data buffer - if NULL only oob data are read/written
 * @oobbuf:	oob data buffer
 *
 * Note, it is allowed to read more than one OOB area at one go, but not write.
 * The interface assumes that the OOB write requests program only one page's
 * OOB area.
 */
struct mtd_oob_ops {
	unsigned int	mode;
	size_t		len;
	size_t		retlen;
	size_t		ooblen;
	size_t		oobretlen;
	uint32_t	ooboffs;
	uint8_t		*datbuf;
	uint8_t		*oobbuf;
};




/* Function Prototypes -------------------------------------------------------*/
void FlashInit(void);

int spi_nand_write(long long to, size_t len, size_t *retlen, const unsigned char *buf);
int spi_nand_write_oob(long long to, struct mtd_oob_ops *ops);
int spi_nand_read(long long from, size_t len, size_t *retlen, unsigned char *buf);
int spi_nand_read_oob(long long from, struct mtd_oob_ops *ops);
int spi_nand_erase(uint64_t addr, uint64_t len);
int spi_nand_init(void);
int spi_nand_internal_data_move(uint32_t page_src, uint32_t page_dst, uint32_t offset, const unsigned char *buf, size_t len);
int spi_nand_lock_tight(void);
int spi_nand_block_markbad(long long offs);
int spi_nand_block_isbad(long long offs);
int spi_nand_lock_block(uint8_t lock);
void spi_nand_release(void);

#endif /* __BATTERY_H */


/*****END OF FILE****/
